/*
  +----------------------------------------------------------------------+
  | PHP Version 5                                                        |
  +----------------------------------------------------------------------+
  | Copyright (c) 1997-2004 The PHP Group                                |
  +----------------------------------------------------------------------+
  | This source file is subject to version 3.0 of the PHP license,       |
  | that is bundled with this package in the file LICENSE, and is        |
  | available through the world-wide-web at the following url:           |
  | http://www.php.net/license/3_0.txt.                                  |
  | If you did not receive a copy of the PHP license and are unable to   |
  | obtain it through the world-wide-web, please send a note to          |
  | license@php.net so we can mail you a copy immediately.               |
  +----------------------------------------------------------------------+
  | Author: Wez Furlong <wez@php.net>                                    |
  | Portions influenced by libevent, written by                          |
  |             Niels Provos <provos@citi.umich.edu>                     |
  +----------------------------------------------------------------------+
*/

/* $Id: php_event_int.h,v 1.3 2004/12/16 03:30:18 wez Exp $ */


#ifndef PHP_EVENT_INT_H
#define PHP_EVENT_INT_H

#ifdef WIN32
# include "win32/time.h"
#endif

#if 0
#if HAVE_SYS_QUEUE_H
# include <sys/queue.h>
#endif
#if HAVE_SYS_TREE_H
# include <sys/tree.h>
#endif
#endif

#ifndef TAILQ_FOREACH
# include "sys_queue.h"
#endif
#ifndef RB_PROTOTYPE
# include "sys_tree.h"
#endif

#if HAVE_SIGACTION
# include <signal.h>
#endif

#define EVLIST_TIMEOUT		0x01
#define EVLIST_INSERTED		0x02
#define EVLIST_SIGNAL		0x04
#define EVLIST_ACTIVE		0x08
#define EVLIST_INTERNAL		0x10
#define EVLIST_INIT			0x80
#define EVLIST_ALL			(0xf000 | 0x9f)

#define EV_TIMEOUT			0x01
#define EV_READ				0x02
#define EV_WRITE			0x04
#define EV_EXCEP			0x08
#define EV_SIGNAL			0x10
#define EV_PERSIST			0x80

#define EVLOOP_ONCE			0x01
#define EVLOOP_NONBLOCK		0x02

#define TIMEOUT_DEFAULT		{5, 0}
	
typedef void (*php_event_callback)(int, short, void *arg TSRMLS_DC);

struct php_event {
#ifdef PHP_WIN32
	OVERLAPPED overlap;
#endif

	TAILQ_ENTRY(php_event)	ev_next;
	TAILQ_ENTRY(php_event)	ev_active_next;
#if HAVE_SIGACTION
	TAILQ_ENTRY(php_event)	ev_signal_next;
#endif
	RB_ENTRY(php_event)		ev_timeout_node;

	int ev_fd;
	short ev_events;
	short ev_ncalls;
	short *ev_pncalls;

	struct timeval ev_timeout;

	int ev_pri; /* smaller == higher pri */

	/* C-based callback */
	php_event_callback ev_callback;
	void *ev_arg;

	/* PHP callback info */
	zval *php_cb;
	zval *php_cb_arg;
	zval *php_stream;
	php_stream *stream;

	int ev_res; /* result passed to callback */
	int ev_flags;

	void *engine_data; /* for use by active engine only */
	struct timeval ev_timeout_interval;
};

TAILQ_HEAD(event_list, php_event);

struct php_eventop {
	char *name;
	void *(*init)(TSRMLS_D);
	void (*fini)(void * TSRMLS_DC);
	int (*add)(void *, struct php_event * TSRMLS_DC);
	int (*del)(void *, struct php_event * TSRMLS_DC);
	int (*recalc)(void *, int TSRMLS_DC);
	int (*dispatch)(void *, struct timeval * TSRMLS_DC);
};

#ifdef PHP_WIN32
extern const struct php_eventop php_event_win32_engine;
#endif
#if HAVE_WORKING_KQUEUE
extern const struct php_eventop php_event_kqueue_engine;
#endif
#if HAVE_EPOLL
extern const struct php_eventop php_event_epoll_engine;
#endif
#if HAVE_SYS_DEVPOLL_H
extern const struct php_eventop php_event_devpoll_engine;
#endif
#if defined(HAVE_SYS_POLL_H) && defined(HAVE_POLL)
extern const struct php_eventop php_event_poll_engine;
#endif
extern const struct php_eventop php_event_select_engine;

ZEND_BEGIN_MODULE_GLOBALS(event)
	const struct php_eventop *evsel;
	void *evbase;

	int event_count;			/* total number of events */
	int event_count_active;		/* total number of active events */

	int got_term;

	/* active event management */
	struct event_list **activequeues;
	int nactivequeues;

	struct event_list eventqueue;
	struct timeval event_tv;

	RB_HEAD(event_tree, php_event) timetree;

#if HAVE_SIGACTION
	struct event_list signalqueue;
	sigset_t evmask;
	int signal_pair[2];
	struct php_event sigevent;
	int signeedsrecalc;
#endif
ZEND_END_MODULE_GLOBALS(event)

#ifdef ZTS
#define EVENT_G(v) TSRMG(event_globals_id, zend_event_globals *, v)
#else
#define EVENT_G(v) (event_globals.v)
#endif

#ifndef timersub
# define timersub(tvp, uvp, vvp)                     \
    do {                                \
        (vvp)->tv_sec = (tvp)->tv_sec - (uvp)->tv_sec;      \
        (vvp)->tv_usec = (tvp)->tv_usec - (uvp)->tv_usec;   \
        if ((vvp)->tv_usec < 0) {               \
            (vvp)->tv_sec--;                \
            (vvp)->tv_usec += 1000000;          \
        }                           \
    } while (0)
#endif
#ifndef timeradd
# define timeradd(tvp, uvp, vvp)                     \
    do {                                \
        (vvp)->tv_sec = (tvp)->tv_sec + (uvp)->tv_sec;      \
        (vvp)->tv_usec = (tvp)->tv_usec + (uvp)->tv_usec;   \
        if ((vvp)->tv_usec >= 1000000) {            \
            (vvp)->tv_sec++;                \
            (vvp)->tv_usec -= 1000000;          \
        }                           \
    } while (0)
#endif

#ifdef PHP_WIN32
/* msvc libc timercmp appears to be broken for <= case, so let's
 * just use the bsd implementation */
# undef timercmp
# define timercmp(tvp, uvp, cmp)                     \
    (((tvp)->tv_sec == (uvp)->tv_sec) ?             \
        ((tvp)->tv_usec cmp (uvp)->tv_usec) :           \
        ((tvp)->tv_sec cmp (uvp)->tv_sec))
#endif

#define event_haveevents()	EVENT_G(event_count)
void event_queue_remove(struct php_event *ev, int queue TSRMLS_DC);
void event_queue_insert(struct php_event *ev, int queue TSRMLS_DC);
int event_add(struct php_event *ev, struct timeval *tv TSRMLS_DC);
int event_del(struct php_event *ev TSRMLS_DC);
void event_active(struct php_event *ev, int res, short ncalls TSRMLS_DC);
void event_set(struct php_event *ev, int fd, short events, php_event_callback callback, void *arg TSRMLS_DC);
int event_pending(struct php_event *ev, short event, struct timeval *tv);

ZEND_DECLARE_MODULE_GLOBALS(event);
#endif

/*
 * Local variables:
 * tab-width: 4
 * c-basic-offset: 4
 * End:
 * vim600: noet sw=4 ts=4 fdm=marker
 * vim<600: noet sw=4 ts=4
 */
